@@ -84,7 +84,9 @@ class ContentNode(object):
         >>> current_content_node.add_child(new_page)
     """
 
-    def __init__(self, document, node_type: str, content="", content_parts=[]):
+    def __init__(self, document, node_type: str, content="", content_parts=None):
+        if content_parts is None:
+            content_parts = []
         self.node_type: str = node_type
         self.content: str = content
         self.document: Document = document
@@ -118,7 +120,7 @@ class ContentNode(object):
         Create a JSON string representation of this ContentNode.
 
             >>> node.to_json()
-    
+
         :return: The JSON formatted string representation of this ContentNode.
         :rtype: str
         """
@@ -176,6 +178,20 @@ class ContentNode(object):
             new_content_node.add_child(ContentNode.from_dict(document, dict_child), dict_child['index'])
         return new_content_node
 
+    def add_child_content(self, node_type, content, index=None):
+        """
+        Convenience method to allow you to quick add a child node with a type and content
+
+        :param node_type: the node type
+        :param content: the content
+        :param index: the index (optional)
+        :return: the new ContentNode
+        """
+        new_node = self.document.create_node(node_type=node_type)
+        new_node.content = content
+        self.add_child(new_node, index)
+        return new_node
+
     def add_child(self, child, index=None):
         """
         Add a ContentNode as a child of this ContentNode
@@ -200,7 +216,7 @@ class ContentNode(object):
         Returns a list of the children of this node.
 
            >>> node.get_children()
-        
+
         :return: The list of child nodes for this ContentNode.
         :rtype: list[ContentNode]
         """
@@ -280,7 +296,7 @@ class ContentNode(object):
            []
 
         :param str feature_type: The type of the feature.
-        
+
         :return: A list of feature with the specified type.  If no features are found, an empty list is returned.
         :rtype: list[ContentFeature]
         """
@@ -295,7 +311,7 @@ class ContentNode(object):
 
         :param str feature_type: The type of the feature.
         :param str name: The name of the feature.
-        
+
         :return: True if the feature is present; else, False.
         :rtype: bool
         """
@@ -373,12 +389,12 @@ class ContentNode(object):
            [ContentNode]
 
         or
-           
+
         >>> document.get_root().select('//*[hasTag($tagName)]', {"tagName": "div"})
            [ContentNode]
 
         :param str selector: The selector (ie. //*)
-        :param variables: A dictionary of variable name/value to use in substituion; defaults to None.  Dictionary keys should match a variable specified in the selector. 
+        :param variables: A dictionary of variable name/value to use in substituion; defaults to None.  Dictionary keys should match a variable specified in the selector.
         :type variables: dict, optional
 
         :return: A list of the matching content nodes.  If no matches are found, the list will be empty.
@@ -392,9 +408,9 @@ class ContentNode(object):
 
     def select_as_node(self, selector, variables=None):
         """
-        Select and return the child nodes of this content node that match the selector value.  
+        Select and return the child nodes of this content node that match the selector value.
         Matching nodes will be returned as the children of a new proxy content node.
-        
+
         Note this doesn't impact this content node's children.  They are not adopted by the proxy node,
         therefore their parents remain intact.
 
@@ -402,12 +418,12 @@ class ContentNode(object):
            ContentNode
 
         or
-           
+
         >>> document.get_root().select_as_node('//*[hasTag($tagName)]', {"tagName": "div"})
            ContentNode
 
         :param str selector: The selector (ie. //*)
-        :param variables: A dictionary of variable name/value to use in substituion; defaults to None.  Dictionary keys should match a variable specified in the selector. 
+        :param variables: A dictionary of variable name/value to use in substituion; defaults to None.  Dictionary keys should match a variable specified in the selector.
         :type variables: dict, optional
 
         :return: A new proxy ContentNode with the matching (selected) nodes as its children.  If no matches are found, the list of children will be empty.
@@ -446,10 +462,10 @@ class ContentNode(object):
         It will be added to the end of the parent
 
             >>> # Get first node of type 'line' from the first page
-            >>> target_child = document.get_root().select('//page')[0].select('//line')[0]  
+            >>> target_child = document.get_root().select('//page')[0].select('//line')[0]
             >>> # Get sixth node of type 'page'
-            >>> target_parent = document.get_root().select('//page')[5]  
-            >>> # Move target_child (line) to the target_parent (sixth page) 
+            >>> target_parent = document.get_root().select('//page')[5]
+            >>> # Move target_child (line) to the target_parent (sixth page)
             >>> document.get_root().move_child_to_parent(target_child, target_parent)
 
         :param ContentNode target_child: The child node that will be moved to a new parent node (target_parent).
@@ -462,7 +478,7 @@ class ContentNode(object):
         """
         This will take a list of content nodes and adopt them under this node, ensuring they are re-parented.
 
-            >>> # select all nodes of type 'line', then the root node 'adopts' them 
+            >>> # select all nodes of type 'line', then the root node 'adopts' them
             >>> # and replaces all it's existing children with these 'line' nodes.
             >>> document.get_root().adopt_children(document.select('//line'), replace=True)
 
@@ -496,7 +512,7 @@ class ContentNode(object):
 
         :param ContentNode end_node: The node to end at
 
-        :return: A list of sibling nodes between this node and the end_node. 
+        :return: A list of sibling nodes between this node and the end_node.
         :rtype: list[ContentNode]
         """
         nodes = []
@@ -554,32 +570,83 @@ class ContentNode(object):
         if start_index is not None:
             [node.tag(tag_to_apply) for node in all_nodes[start_index:end_index]]
 
+    def tag_text_tree(self):
+        """
+        Return a text tree
+        :return:
+        """
+        from anytree import RenderTree
+        result = ""
+        for pre, _, node in RenderTree(self):
+            result = result + ("%s%s" % (pre, f"{node.content} ({node.node_type}) {node.get_tags()}\n"))
+        return result
+
     def tag(self, tag_to_apply, selector=".", content_re=None,
-            use_all_content=False, node_only=False,
-            fixed_position=None, data=None):
+            use_all_content=False, node_only=None,
+            fixed_position=None, data=None, separator=" "):
         """
         This will tag (see Feature Tagging) the expression groups identified by the regular expression.
 
             >>> document.content_node.tag('is_cheese')
 
+        Note that if you use the flag use_all_content then node_only will default to True if not set, else it
+        will default to False
+
         :param tag_to_apply: the name of tag that will be applied to the node
         :param selector: The selector to identify the source nodes to work on (default . - the current node)
         :param content_re: the regular expression that you wish to use to tag, note that we will create a tag for each matching group
         :param use_all_content: apply the regular expression to the all_content (include content from child nodes)
+        :param separator: Separator to use for use_all_content
         :param node_only: Ignore the matching groups and tag the whole node
         :param fixed_position: use a fixed position, supplied as a tuple i.e. - (4,10) tag from position 4 to 10 (default None)
         :param data: Attach the a dictionary of data for the given tag
+
         """
 
+        if use_all_content and node_only is None:
+            node_only = True
+        elif node_only is None:
+            node_only = False
+
+        def tag_node_position(node_to_check, start, end, node_data):
+
+            # Make sure we have content on the node
+            if node_to_check.content:
+                if len(node_to_check.content) > 0:
+                    if start < len(node_to_check.content) and end < len(node_to_check.content):
+                        node_to_check.add_feature('tag', tag_to_apply,
+                                                  Tag(start, end,
+                                                      node_to_check.content[start:end],
+                                                      data=node_data))
+                        return -1
+                    elif start < len(node_to_check.content) <= end:
+                        node_to_check.add_feature('tag', tag_to_apply,
+                                                  Tag(start,
+                                                      len(node_to_check.content),
+                                                      value=node_to_check.content[start:],
+                                                      data=node_data))
+
+                end = end - len(node_to_check.content) + len(separator)
+                start = 0 if start - len(node_to_check.content) - len(separator) < 0 else start - len(
+                    node_to_check.content) - len(separator)
+
+            for child_node in node_to_check.children:
+                result = tag_node_position(child_node, start, end, node_data)
+                if result < 0:
+                    return -1
+                else:
+                    end = end - result
+                    start = 0 if start - result < 0 else start - result
+
+            return end
+
         if content_re:
             pattern = re.compile(content_re)
 
         for node in self.select(selector):
             if fixed_position:
-                node.add_feature('tag', tag_to_apply,
-                                 Tag(fixed_position[0], fixed_position[1],
-                                     node.content[fixed_position[0]:fixed_position[1]],
-                                     data))
+                tag_node_position(node, fixed_position[0], fixed_position[1], data)
+
             else:
                 if not content_re:
                     node.add_feature('tag', tag_to_apply, Tag(data=data))
@@ -590,7 +657,7 @@ class ContentNode(object):
                         else:
                             content = None
                     else:
-                        content = node.get_all_content()
+                        content = node.get_all_content(separator=separator)
 
                     if content is not None:
                         match = pattern.match(content)
@@ -600,8 +667,17 @@ class ContentNode(object):
                             else:
                                 for index, m in enumerate(match.groups()):
                                     idx = index + 1
-                                    node.add_feature('tag', tag_to_apply,
-                                                     Tag(match.start(idx), match.end(idx), match.group(idx), data=data))
+
+                                    if node_only:
+                                        node.add_feature('tag', tag_to_apply,
+                                                         Tag(match.start(idx), match.end(idx), match.group(idx),
+                                                             data=data))
+                                    else:
+                                        # We need to work out where the content is in the child nodes
+                                        start_offset = match.start(idx)
+                                        end_offset = match.end(idx)
+
+                                        tag_node_position(node, start_offset, end_offset, data)
 
     def get_tags(self):
         """
@@ -614,6 +690,24 @@ class ContentNode(object):
         """
         return [i.name for i in self.get_features_of_type("tag")]
 
+    def get_tag_values(self, tag_name, include_children=False):
+        """
+        Get the values for a specific tag name
+
+        :param tag_name: tag name
+        :param include_children: include the children of this node
+        :return: a list of the tag values
+        """
+        values = []
+        for tag in self.get_tag(tag_name):
+            values.append(tag.value)
+
+        if include_children:
+            for child in self.get_children():
+                values.extend(child.get_tag_values(tag_name, include_children))
+
+        return values
+
     def get_tag(self, tag_name):
         """
         Returns the value of a tag, this can be either a single list [start,end,value] or if multiple parts of the
@@ -624,9 +718,17 @@ class ContentNode(object):
 
         :param tag_name: The name of the tag
 
-        :return: The tagged location and value (or a list if more than one)
+        :return: A list tagged location and values for this label in this node
         """
-        return self.get_feature_value('tag', tag_name)
+        tag_details = self.get_feature_value('tag', tag_name)
+
+        if tag_details is None:
+            return []
+
+        if isinstance(tag_details, list):
+            return tag_details
+        else:
+            return [tag_details]
 
     def get_all_tags(self):
         """
@@ -762,7 +864,7 @@ class ContentNode(object):
     def is_first_child(self):
         """
         Determines if this node is the first child of its parent or has no parent.
-        
+
         :return: True if this node is the first child of its parent or if this node has no parent; else, False;
         :rtype: bool
         """
@@ -805,9 +907,9 @@ class ContentNode(object):
     def get_node_at_index(self, index):
         """
         Returns the child node at the specified index. If the specified index is outside the first (0), or
-        last child's index, None is returned.  
-        
-        Note:  documents allow for sparse representation and child nodes may not have consecutive index numbers. 
+        last child's index, None is returned.
+
+        Note:  documents allow for sparse representation and child nodes may not have consecutive index numbers.
         If there isn't a child node at the specfied index, a 'virtual' node will be returned.  This 'virtual' node
         will have the node type of its nearest sibling and will have an index value, but will have no features or content.
 
@@ -873,10 +975,10 @@ class ContentNode(object):
 
     def next_node(self, node_type_re='.*', skip_virtual=False, has_no_content=True):
         """
-        Returns the next sibling content node. 
-        
+        Returns the next sibling content node.
+
         Note:  This logic relies on node indexes.  Documents allow for sparse representation and child nodes may not have consecutive index numbers.
-        Therefore, the next node might actually be a virtual node that is created to fill a gap in the document.  You can skip virtual nodes by setting the 
+        Therefore, the next node might actually be a virtual node that is created to fill a gap in the document.  You can skip virtual nodes by setting the
         skip_virtual parameter to False.
 
         :param node_type_re: The regular expression to match against the next sibling node's type; default is '.*'.
@@ -906,10 +1008,10 @@ class ContentNode(object):
 
     def previous_node(self, node_type_re='.*', skip_virtual=False, has_no_content=False, traverse=Traverse.SIBLING):
         """
-        Returns the previous sibling content node. 
-        
+        Returns the previous sibling content node.
+
         Note:  This logic relies on node indexes.  Documents allow for sparse representation and child nodes may not have consecutive index numbers.
-        Therefore, the previous node might actually be a virtual node that is created to fill a gap in the document.  You can skip virtual nodes by setting the 
+        Therefore, the previous node might actually be a virtual node that is created to fill a gap in the document.  You can skip virtual nodes by setting the
         skip_virtual parameter to False.
 
         :param node_type_re: The regular expression to match against the previous node's type; default is '.*'.
@@ -1140,13 +1242,13 @@ class Document(object):
         return msgpack.packb(self.to_dict(), use_bin_type=True)
 
     def to_json(self):
-        """ 
+        """
         Create a JSON string representation of this Document.
 
             >>> document.to_json()
 
         :return: The JSON formatted string representation of this Document.
-        :rtype: str 
+        :rtype: str
         """
         return json.dumps(self.to_dict(), ensure_ascii=False)
 
@@ -1169,7 +1271,7 @@ class Document(object):
             >>> document.to_dict()
 
         :return: A dictionary representation of this Document.
-        :rtype: dict  
+        :rtype: dict
         """
         return {'version': Document.CURRENT_VERSION, 'metadata': self.metadata,
                 'content_node': self.content_node.to_dict() if self.content_node else None,
@@ -1186,7 +1288,7 @@ class Document(object):
             >>> Document.from_dict(doc_dict)
 
         :param dict doc_dict: A dictionary representation of a Kodexa Document.
-        
+
         :return: A complete Kodexa Document
         :rtype: Document
         """
@@ -1212,7 +1314,7 @@ class Document(object):
             >>> Document.from_json(json_string)
 
         :param str json_string: A JSON string representation of a Kodexa Document
-        
+
         :return: A complete Kodexa Document
         :rtype: Document
         """
@@ -1255,7 +1357,7 @@ class Document(object):
     def create_node(self, node_type: str, content: str = None, virtual: bool = False, parent: ContentNode = None,
                     index: int = 0):
         """
-        Creates a new node for the document.  The new node is not added to the document, but any mixins that have been 
+        Creates a new node for the document.  The new node is not added to the document, but any mixins that have been
         applied to the document will also be available on the new node.
 
             >>> document.create_node(node_type='page')
@@ -1264,7 +1366,7 @@ class Document(object):
 
         :param str node_type: The type of node.
         :param str content: The content for the node; defaults to None.
-        :param bool virtual: Indicates if this is a 'real' or 'virtual' node; default is False.  'Real' nodes contain document content.  
+        :param bool virtual: Indicates if this is a 'real' or 'virtual' node; default is False.  'Real' nodes contain document content.
         'Virtual' nodes are synthesized as necessary to fill gaps in between non-consecutively indexed siblings.  Such indexing arises when document content is sparse.
         :param ContentNode parent: The parent for this newly created node; default is None;
         :param int index: The index property to be set on this node; default is 0;
@@ -1292,7 +1394,7 @@ class Document(object):
         Creates a Document that has a 'file-handle' connector to the specified file.
 
         :param file file: The file to which the new Document is connected.
-        
+
         :return: A Document connected to the specified file.
         :rtype: Document
         """
@@ -1307,8 +1409,8 @@ class Document(object):
         Creates a Document that has a 'url' connector for the specified url.
 
         :param str url: The URL to which the new Document is connected.
-        :param dict headers: Headers that should be used when reading from the URL 
-        
+        :param dict headers: Headers that should be used when reading from the URL
+
         :return: A Document connected to the specified URL with the specified headers (if any).
         :rtype: Document
         """
@@ -1328,8 +1430,8 @@ class Document(object):
            [ContentNode]
 
         :param str selector: The selector (ie. //*)
-        :param variables: A dictionary of variable name/value to use in substituion; defaults to an empty dictionary.  Dictionary keys should match a variable specified in the selector. 
-        :type variables: dict, optional 
+        :param variables: A dictionary of variable name/value to use in substituion; defaults to an empty dictionary.  Dictionary keys should match a variable specified in the selector.
+        :type variables: dict, optional
 
         :return: A list of the matching ContentNodes.  If no matches found, list is empty.
         :rtype: list[ContentNodes]
@@ -1379,7 +1481,7 @@ class KodexaRender:
     def build_html(self, document: Document, node: ContentNode):
         render_uuid = str(uuid.uuid4())
         return """
-  <div id='kodexa-div-""" + render_uuid + """'></div> 
+  <div id='kodexa-div-""" + render_uuid + """'></div>
   <script>
 
 require.config({
@@ -1395,7 +1497,7 @@ require(['kodexa-lib-""" + render_uuid + """','jquery'], function() {
        widget.attach($('#kodexa-div-""" + render_uuid + """'))
        widget.render(""" + ("'" + node.uuid + "'" if node else "") + """);
     });
-    
+
 });
 </script>
 """
