@@ -187,29 +187,51 @@ class PipelineStep:
     It is internally used by the Pipeline and is not a public API
     """
 
-    def __init__(self, step, enabled=False, condition=None, name=None):
+    def __init__(self, step, enabled=False, condition=None, name=None, options=None, attach_source=False,
+                 parameterized=False):
+        if options is None:
+            options = {}
         self.step = step
         self.name = name
         self.condition = condition
         self.enabled = enabled
+        self.options = options
+        self.parameterized = parameterized
 
-        if callable(self.step):
+        if str(type(self.step)) == "<class 'type'>":
+            logging.info(f"Adding new step class {step.__name__} to pipeline {self.name}")
+            self.step = step
+        elif callable(self.step):
             logging.info(f"Adding new step function {step.__name__} to pipeline {self.name}")
             self.name = step.__name__
+        elif isinstance(self.step, str):
+            logging.info(f"Adding new remote step {step} to pipeline {self.name}")
+            from kodexa import RemoteAction
+            self.step = RemoteAction(step, options=options, attach_source=attach_source)
         else:
             logging.info(f"Adding new step {step.get_name()} to pipeline {self.name}")
 
     def to_dict(self):
         try:
-            if callable(self.step):
+            if str(type(self.step)) == "<class 'type'>":
+                raise Exception("You can not yet deploy a pipeline with a class instance style step")
+            elif isinstance(self.step, str):
+                return {
+                    'ref': self.step,
+                    'options': self.options
+                }
+            elif callable(self.step):
                 metadata = {
                     'function': self.step.__name__,
                     'script': dedent(inspect.getsource(self.step))
                 }
             else:
                 metadata = self.step.to_dict()
+
             metadata['name'] = self.name
             metadata['condition'] = self.condition
+            metadata['conditional'] = self.condition is not None
+            metadata['parameterized'] = self.parameterized
             metadata['enabled'] = self.enabled
             return metadata
         except AttributeError as e:
@@ -220,8 +242,43 @@ class PipelineStep:
             try:
 
                 context.set_current_document(document)
+                if str(type(self.step)) == "<class 'type'>":
 
-                if not callable(self.step):
+                    # We need to handle the parameterization
+                    import copy
+
+                    option_copy = copy.deepcopy(self.options)
+                    if self.parameterized:
+                        import collections
+
+                        def replace_params(opts, params):
+                            if isinstance(opts, dict):
+                                for key, val in opts.items():
+                                    opts[key] = replace_params(val, params)
+                            elif isinstance(opts, list):
+                                new_list = []
+                                for list_val in opts:
+                                    new_list.append(replace_params(list_val))
+                                return new_list
+                            elif isinstance(opts, str):
+                                if opts.startswith('${') and opts.endswith('}'):
+                                    param_name = opts[2:-1]
+                                    if param_name in params:
+                                        return params[param_name]
+                                    else:
+                                        raise Exception("Unable to find parameter name " + param_name)
+
+                            return opts
+
+                        replace_params(option_copy, context.parameters)
+
+                    step_instance = self.step(**option_copy)
+                    if len(signature(step_instance.process).parameters) == 1:
+                        return step_instance.process(document)
+                    else:
+                        return step_instance.process(document, context)
+
+                elif not callable(self.step):
                     logging.info(f"Starting step {self.step.get_name()}")
 
                     if len(signature(self.step.process).parameters) == 1:
@@ -254,8 +311,10 @@ class PipelineStep:
 
         if self.condition:
             from simpleeval import simple_eval
+            from addict import Dict
+            addict_dict = Dict(context.context)
             return bool(
-                simple_eval(self.condition, names={'document': document, 'step': self}))
+                simple_eval(self.condition, names={'context': addict_dict}))
 
         return True
 
@@ -304,7 +363,8 @@ class Pipeline:
 
         return self
 
-    def add_step(self, step, name=None, enabled=True, condition=None):
+    def add_step(self, step, name=None, enabled=True, condition=None, options=None, attach_source=False,
+                 parameterized=False):
         """
         Add the given step to the current pipeline
 
@@ -318,12 +378,24 @@ class Pipeline:
             >>>      return doc
             >>> pipeline.add_step(my_function)
 
+        If you are using remote actions on a server, or for deployment to a remote
+        pipeline you can also use a shorthand
+
+            >>> pipeline.add_step('kodexa/html-parser',options={'summarize':False})
+
         :param step: the step to add
         :param name: the name to use to describe the step (default None)
         :param enabled: is the step enabled (default True)
         :param condition: condition to evaluate before executing the step (default None)
+        :param options: options to be passed to the step if it is a simplified remote action
+        :param attach_source: if step is simplified remote action this determines if we need to add the source
+        :param parameterized: apply the pipeline's parameters to the options
+
         """
-        self.steps.append(PipelineStep(step=step, name=name, enabled=enabled, condition=condition))
+        if options is None:
+            options = {}
+        self.steps.append(PipelineStep(step=step, name=name, enabled=enabled, condition=condition, options=options,
+                                       attach_source=attach_source, parameterized=parameterized))
 
         return self
 
@@ -362,7 +434,7 @@ class Pipeline:
 
         return yaml.dump(configuration_steps)
 
-    def run(self):
+    def run(self, parameters=None):
         """
         Run the current pipeline, note that you must have a sink in place to allow the pipeline to run
 
@@ -372,7 +444,12 @@ class Pipeline:
 
         :return: The context from the run
         """
+        if parameters is None:
+            parameters = {}
+
         self.context.statistics = PipelineStatistics()
+        self.context.parameters = parameters
+
         logging.info(f"Starting pipeline {self.name}")
         for document in self.connector:
 
@@ -382,24 +459,26 @@ class Pipeline:
             for step in self.steps:
                 document = step.execute(self.context, document)
 
-            if self.sink:
-                logging.info(f"Writing to sink {self.sink.get_name()}")
-                try:
-                    self.sink.sink(document)
-                except:
-                    if document:
-                        document.exceptions.append({
-                            "step": self.sink.get_name(),
-                            "exception": sys.exc_info()[0]
-                        })
-                    if self.context.stop_on_exception:
-                        raise
-
             if document:
+                if self.sink:
+                    logging.info(f"Writing to sink {self.sink.get_name()}")
+                    try:
+                        self.sink.sink(document)
+                    except:
+                        if document:
+                            document.exceptions.append({
+                                "step": self.sink.get_name(),
+                                "exception": sys.exc_info()[0]
+                            })
+                        if self.context.stop_on_exception:
+                            raise
+
                 document.log = log_stream.getvalue()
 
-            self.context.statistics.processed_document(document)
-            self.context.output_document = document
+                self.context.statistics.processed_document(document)
+                self.context.output_document = document
+            else:
+                logging.warning("A step did not return a document?")
 
         logging.info(f"Completed pipeline {self.name}")
 
@@ -487,5 +566,5 @@ class PipelineStatistics:
         """
         self.documents_processed += 1
 
-        if document.exceptions:
+        if document and document.exceptions:
             self.document_exceptions += 1
